import io
import math
import re

from rubymarshal.classes import (
    Module,
    RubyObject,
    RubyString,
    Symbol,
    UserDef,
    UsrMarshal,
)
from rubymarshal.constants import (
    TYPE_ARRAY,
    TYPE_BIGNUM,
    TYPE_CLASS,
    TYPE_FALSE,
    TYPE_FIXNUM,
    TYPE_FLOAT,
    TYPE_HASH,
    TYPE_IVAR,
    TYPE_LINK,
    TYPE_MODULE,
    TYPE_NIL,
    TYPE_OBJECT,
    TYPE_REGEXP,
    TYPE_STRING,
    TYPE_SYMBOL,
    TYPE_SYMLINK,
    TYPE_TRUE,
    TYPE_USERDEF,
    TYPE_USRMARSHAL,
)
from rubymarshal.utils import write_sbyte, write_ubyte, write_ushort

__author__ = "Matthieu Gallet"

re_class = re.compile("").__class__
simple_float_re = re.compile(r"^\d+\.\d*0+$")


class Writer:
    def __init__(self, fd):
        self.symbols = {}
        self.objects = {}
        self.fd = fd

    def write(self, obj):
        if obj is None:
            self.write_none()
        elif obj is False:
            self.write_false()
        elif obj is True:
            self.write_true()
        elif isinstance(obj, int):
            self.write_int(obj)
        elif isinstance(obj, Symbol):
            self.write_symbol(obj)
        elif isinstance(obj, list):
            self.write_list(obj)
        elif isinstance(obj, dict):
            self.write_dict(obj)
        elif isinstance(obj, bytes):
            self.write_bytes(obj)
        elif isinstance(obj, str):
            self.write_string(obj)
        elif isinstance(obj, RubyString):
            self.write_ruby_string(obj)
        elif isinstance(obj, float):
            self.write_float(obj)
        elif isinstance(obj, re_class):
            self.write_regexp(obj)
        elif isinstance(obj, Module):
            self.write_module(obj)
        elif isinstance(obj, UsrMarshal):
            self.write_usr_marshal(obj)
        elif isinstance(obj, UserDef):
            self.write_user_def(obj)
        elif isinstance(obj, RubyObject):
            self.write_ruby_object(obj)
        elif isinstance(obj, type) and issubclass(obj, RubyObject):
            self.write_class(obj)
        else:
            self.write_python_object(obj)

    def write_python_object(self, obj):
        """override this method to dump new Python classes"""
        raise ValueError("unmarshable object: %s(%r)" % (obj.__class__.__name__, obj))

    def write_true(self):
        self.fd.write(TYPE_TRUE)

    def write_false(self):
        self.fd.write(TYPE_FALSE)

    def write_none(self):
        self.fd.write(TYPE_NIL)

    def write_class(self, obj):
        self.fd.write(TYPE_CLASS)
        self.write_long(len(obj.ruby_class_name.encode()))
        self.fd.write(obj.ruby_class_name.encode())

    def write_ruby_object(self, obj):
        if self.must_write(obj):
            self.fd.write(TYPE_OBJECT)
            self.write(Symbol(obj.ruby_class_name))
            if not isinstance(obj.attributes, dict):
                raise ValueError("%r values is not a dict" % obj)
            self.write_attributes(obj.attributes)

    def write_user_def(self, obj):
        if self.must_write(obj):
            if obj.attributes:
                self.fd.write(TYPE_IVAR)
            self.fd.write(TYPE_USERDEF)
            self.write(Symbol(obj.ruby_class_name))
            # noinspection PyProtectedMember
            bdata = obj._dump()
            self.write_long(len(bdata))
            self.fd.write(bdata)
            if obj.attributes:
                self.write_attributes(obj.attributes)

    def write_usr_marshal(self, obj):
        if self.must_write(obj):
            if obj.attributes:
                self.fd.write(TYPE_IVAR)
            self.fd.write(TYPE_USRMARSHAL)
            self.write(Symbol(obj.ruby_class_name))
            private_data = obj.marshal_dump()
            self.write(private_data)
            if obj.attributes:
                self.write_attributes(obj.attributes)

    def write_module(self, obj):
        self.fd.write(TYPE_MODULE)
        self.write_long(len(obj.ruby_class_name.encode()))
        self.fd.write(obj.ruby_class_name.encode())

    def write_regexp(self, obj):
        flags = 0
        if obj.flags & re.IGNORECASE:
            flags += 1
        if obj.flags & re.MULTILINE:
            flags += 4
        self.fd.write(TYPE_IVAR)
        self.fd.write(TYPE_REGEXP)
        pattern = obj.pattern.encode("utf-8")
        self.write_long(len(pattern))
        self.fd.write(pattern)
        write_ubyte(self.fd, flags)
        self.write_long(1)
        self.write(Symbol("E"))
        self.write(False)

    def write_float(self, obj):
        obj = "%.20g" % obj
        if simple_float_re.match(obj):
            while obj.endswith("0"):
                obj = obj[:-1]
        obj = obj.encode("utf-8")
        self.fd.write(TYPE_FLOAT)
        self.write_long(len(obj))
        self.fd.write(obj)

    def write_ruby_string(self, obj):
        if self.must_write(obj):
            encoding = "utf-8"
            attributes = obj.attributes
            if "E" in attributes and not attributes["E"]:
                encoding = "latin-1"
            elif "encoding" in attributes:
                encoding = attributes["encoding"].decode()
            else:
                attributes["E"] = True
            encoded = obj.encode(encoding)
            self.fd.write(TYPE_IVAR)
            self.write_bytes(encoded)
            self.write_attributes(attributes)

    def write_string(self, obj):
        obj = obj.encode("utf-8")
        self.fd.write(TYPE_IVAR)
        self.write_bytes(obj)
        self.write_long(1)
        self.write(Symbol("E"))
        self.write(True)

    def write_bytes(self, obj):
        self.fd.write(TYPE_STRING)
        self.write_long(len(obj))
        self.fd.write(obj)

    def write_dict(self, obj):
        if self.must_write(obj):
            self.fd.write(TYPE_HASH)
            self.write_long(len(obj))
            for key, value in obj.items():
                self.write(key)
                self.write(value)

    def write_list(self, obj):
        if self.must_write(obj):
            self.fd.write(TYPE_ARRAY)
            self.write_long(len(obj))
            for x in obj:
                self.write(x)

    def write_symbol(self, obj):
        if obj.name in self.symbols:
            self.fd.write(TYPE_SYMLINK)
            self.write_long(self.symbols[obj.name])
        else:
            self.fd.write(TYPE_SYMBOL)
            symbol_index = len(self.symbols)
            self.symbols[obj.name] = symbol_index
            encoded = obj.name.encode("utf-8")
            self.write_long(len(encoded))
            self.fd.write(encoded)

    def write_int(self, obj):
        if obj.bit_length() <= 5 * 8:
            self.fd.write(TYPE_FIXNUM)
            # noinspection PyTypeChecker
            self.write_long(obj)
        else:
            self.fd.write(TYPE_BIGNUM)
            if obj < 0:
                self.fd.write(b"-")
            else:
                self.fd.write(b"+")
            obj = abs(obj)
            size = int(math.ceil(obj.bit_length() / 16.0))
            self.write_long(size)
            for i in range(size):
                self.write_short(obj % 65536)
                obj //= 65536

    def write_attributes(self, attributes):
        self.write_long(len(attributes))
        for attr_name, attr_value in attributes.items():
            self.write(Symbol(attr_name))
            self.write(attr_value)

    def write_short(self, obj):
        write_ushort(self.fd, obj)

    def write_long(self, obj):
        if obj == 0:
            self.fd.write(b"\0")
        elif 0 < obj < 123:
            write_sbyte(self.fd, obj + 5)
        elif -124 < obj < 0:
            write_sbyte(self.fd, obj - 5)
        else:
            size = int(math.ceil(obj.bit_length() / 8.0))
            if size > 5:
                raise ValueError("%d too long for serialization" % obj)
            original_obj = obj
            factor = 256**size
            if obj < 0 and obj == -factor:
                size -= 1
                obj += factor / 256
            elif obj < 0:
                obj += factor
            sign = int(math.copysign(size, original_obj))
            write_sbyte(self.fd, sign)
            for i in range(size):
                write_ubyte(self.fd, obj % 256)
                obj //= 256

    def must_write(self, obj):
        """return False if the object has already been serialized (and write a link to it),
        otherwise return True"""
        if id(obj) in self.objects:
            self.fd.write(TYPE_LINK)
            self.write_long(self.objects[id(obj)])
            return False
        else:
            link_index = len(self.objects)
            self.objects[id(obj)] = link_index
            return True


def write(fd, obj, cls=Writer):
    """write an Python object to a file descriptor

    :param fd: the file descriptor
    :param obj: the object to serialize
    :param cls: Writer class to use. Subclass it to serialize new Python classes
    """
    fd.write(b"\x04\x08")
    writer = cls(fd)
    writer.write(obj)


def writes(obj, cls=Writer):
    """write an Python object to a bytes string

    :param obj: the object to serialize
    :param cls: Writer class to use. Subclass it to serialize new Python classes
    """
    fd = io.BytesIO()
    write(fd, obj, cls=cls)
    return fd.getvalue()



from typing import Type

__author__ = "Matthieu Gallet"


class RubyObject:
    ruby_class_name = None

    def __init__(self, ruby_class_name=None, attributes=None):
        self.ruby_class_name = ruby_class_name or self.ruby_class_name
        self.attributes = attributes or {}

    def set_attributes(self, attributes):
        self.attributes = attributes

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.attributes == other.attributes

    def __hash__(self):
        return hash(hash(self.__class__.__name__) + hash(repr(self.attributes)))

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.attributes)

    def __str__(self):
        return "%s(%r)" % (self.__class__.__name__, self.attributes)


class RubyString(RubyObject):
    def __init__(self, text: str, attributes=None):
        self.text = text
        super().__init__("str", attributes=attributes)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.text == other
        elif isinstance(other, RubyString):
            return self.text == other.text and self.attributes == other.attributes
        return False

    def __ne__(self, other):
        if isinstance(other, str):
            return self.text != other
        elif isinstance(other, RubyString):
            return self.text != other.text or self.attributes != other.attributes
        return False

    def __getattr__(self, item):
        return getattr(self.text, item)

    def __add__(self, other):
        return RubyString(self.text + str(other), self.attributes)

    def __hash__(self):
        return hash(self.text)

    def __repr__(self):
        return repr(self.text)

    def __str__(self):
        return self.text

    def __lt__(self, other):
        return self.text < other

    def __gt__(self, other):
        return self.text > other

    def __le__(self, other):
        return self.text <= other

    def __ge__(self, other):
        return self.text >= other

    def __iter__(self):
        yield from self.text

    def __bool__(self):
        return bool(self.text)

    def __getitem__(self, item):
        return self.text[item]

    def __len__(self):
        return len(self.text)


class UsrMarshal(RubyObject):
    """object with a user-defined serialization format using the marshal_dump and marshal_load instance methods.
    Upon loading a new instance must be allocated and marshal_load must be called on the instance with the data."""

    def __init__(self, ruby_class_name=None, attributes=None):
        self._private_data = None
        super().__init__(ruby_class_name=ruby_class_name, attributes=attributes)

    def marshal_load(self, private_data):
        self._private_data = private_data

    def marshal_dump(self):
        return self._private_data


class UserDef(RubyObject):
    """object with a user-defined serialization format using the _dump instance method and _load class method.

    data is a byte sequence containing the user-defined representation of the object.

    The class method _load is called on the class with a string created from the byte-sequence."""

    def __init__(self, ruby_class_name=None, attributes=None):
        self._private_data = None
        super().__init__(ruby_class_name=ruby_class_name, attributes=attributes)

    def _load(self, private_data: bytes):
        self._private_data = private_data

    def _dump(self) -> bytes:
        return self._private_data


class Extended(RubyObject):
    pass


class Module(RubyObject):
    pass


class Symbol:
    __registered_symbols__ = {}

    def __new__(cls, name):
        if name in cls.__registered_symbols__:
            return cls.__registered_symbols__[name]
        return super(Symbol, cls).__new__(cls)

    def __init__(self, name):
        self.name = name
        self.__registered_symbols__[name] = self

    def __hash__(self):
        return hash("<<<:%s:>>>" % self.name)

    def __repr__(self):
        return 'Symbol("%s")' % self.name

    def __str__(self):
        return ":%s" % self.name

    def encode(self, *args, **kwargs):
        return self.name.encode(*args, **kwargs)


class ClassRegistry:
    def __init__(self):
        self._registry = {}

    def register(self, cls: Type[RubyObject]):
        assert issubclass(cls, RubyObject)
        self._registry[cls.ruby_class_name] = cls

    def unregister(self, cls: Type[RubyObject]):
        assert issubclass(cls, RubyObject)
        if cls.ruby_class_name in self._registry:
            del self._registry[cls.ruby_class_name]

    def get(self, ruby_class_name: str, default_cls: Type[RubyObject]):
        return self._registry.get(ruby_class_name, default_cls)

    def __contains__(self, item):
        return item in self._registry

    def __getitem__(self, item):
        return self._registry[item]

    def __delitem__(self, key):
        del self._registry[key]


registry = ClassRegistry()



"""regroup all struct functions"""
import struct

__author__ = "Matthieu Gallet"


def write_ushort(fd, obj):
    fd.write(struct.pack("<H", obj))


def write_sbyte(fd, obj):
    fd.write(struct.pack("b", obj))


def write_ubyte(fd, obj):
    fd.write(struct.pack("B", obj))


def read_ushort(fd):
    return struct.unpack("<H", fd.read(2))[0]


def read_sbyte(fd):
    return struct.unpack("b", fd.read(1))[0]


def read_ubyte(fd):
    return struct.unpack("B", fd.read(1))[0]



__author__ = "Matthieu Gallet"

TYPE_NIL = b"0"
TYPE_TRUE = b"T"
TYPE_FALSE = b"F"
TYPE_FIXNUM = b"i"
TYPE_ARRAY = b"["
TYPE_SYMBOL = b":"
TYPE_STRING = b'"'
TYPE_IVAR = b"I"
TYPE_HASH = b"{"
TYPE_FLOAT = b"f"
TYPE_BIGNUM = b"l"
TYPE_REGEXP = b"/"
TYPE_USRMARSHAL = b"U"
TYPE_SYMLINK = b";"
TYPE_LINK = b"@"
TYPE_EXTENDED = b"e"
TYPE_UCLASS = b"C"
TYPE_OBJECT = b"o"
TYPE_DATA = b"d"
TYPE_USERDEF = b"u"
TYPE_HASH_DEF = b"}"
TYPE_STRUCT = b"S"
TYPE_MODULE_OLD = b"M"
TYPE_CLASS = b"c"
TYPE_MODULE = b"m"

TYPES = {
    TYPE_NIL: "TYPE_NIL",
    TYPE_TRUE: "TYPE_TRUE",
    TYPE_FALSE: "TYPE_FALSE",
    TYPE_FIXNUM: "TYPE_FIXNUM",
    TYPE_ARRAY: "TYPE_ARRAY",
    TYPE_SYMBOL: "TYPE_SYMBOL",
    TYPE_STRING: "TYPE_STRING",
    TYPE_IVAR: "TYPE_IVAR",
    TYPE_HASH: "TYPE_HASH",
    TYPE_FLOAT: "TYPE_FLOAT",
    TYPE_BIGNUM: "TYPE_BIGNUM",
    TYPE_REGEXP: "TYPE_REGEXP",
    TYPE_USRMARSHAL: "TYPE_USRMARSHAL",
    TYPE_SYMLINK: "TYPE_SYMLINK",
    TYPE_LINK: "TYPE_LINK",
    TYPE_EXTENDED: "TYPE_EXTENDED",
    TYPE_UCLASS: "TYPE_UCLASS",
    TYPE_OBJECT: "TYPE_OBJECT",
    TYPE_DATA: "TYPE_DATA",
    TYPE_USERDEF: "TYPE_USERDEF",
    TYPE_HASH_DEF: "TYPE_HASH_DEF",
    TYPE_STRUCT: "TYPE_STRUCT",
    TYPE_MODULE_OLD: "TYPE_MODULE_OLD",
    TYPE_CLASS: "TYPE_CLASS",
    TYPE_MODULE: "TYPE_MODULE",
}



import io
import re

from rubymarshal.classes import (
    Extended,
    Module,
    RubyObject,
    RubyString,
    Symbol,
    UserDef,
    UsrMarshal,
)
from rubymarshal.classes import registry as global_registry
from rubymarshal.constants import (
    TYPE_ARRAY,
    TYPE_BIGNUM,
    TYPE_CLASS,
    TYPE_DATA,
    TYPE_EXTENDED,
    TYPE_FALSE,
    TYPE_FIXNUM,
    TYPE_FLOAT,
    TYPE_HASH,
    TYPE_IVAR,
    TYPE_LINK,
    TYPE_MODULE,
    TYPE_NIL,
    TYPE_OBJECT,
    TYPE_REGEXP,
    TYPE_STRING,
    TYPE_STRUCT,
    TYPE_SYMBOL,
    TYPE_SYMLINK,
    TYPE_TRUE,
    TYPE_USERDEF,
    TYPE_USRMARSHAL,
)
from rubymarshal.utils import read_sbyte, read_ubyte, read_ushort

__author__ = "Matthieu Gallet"


class Reader:
    def __init__(self, fd, registry=None):
        self.symbols = []
        self.objects = []
        self.fd = fd
        self.registry = registry or global_registry

    def read(self, token=None):
        if token is None:
            token = self.fd.read(1)

        # From https://docs.ruby-lang.org/en/2.1.0/marshal_rdoc.html:
        # The stream contains only one copy of each object for all objects except
        # true, false, nil, Fixnums and Symbols.
        object_index = None
        if token in (
            TYPE_IVAR,
            # TYPE_EXTENDED, TYPE_UCLASS, ????
            TYPE_CLASS,
            TYPE_MODULE,
            TYPE_FLOAT,
            TYPE_BIGNUM,
            TYPE_REGEXP,
            TYPE_ARRAY,
            TYPE_HASH,
            TYPE_STRUCT,
            TYPE_OBJECT,
            TYPE_DATA,
            TYPE_USRMARSHAL,
        ):
            self.objects.append(None)
            object_index = len(self.objects)

        result = None
        if token == TYPE_NIL:
            pass
        elif token == TYPE_TRUE:
            result = True
        elif token == TYPE_FALSE:
            result = False
        elif token == TYPE_IVAR:
            sub_token = self.fd.read(1)
            result = self.read(sub_token)
            flags = None
            if sub_token == TYPE_REGEXP:
                options = ord(self.fd.read(1))
                flags = 0
                if options & 1:
                    flags |= re.IGNORECASE
                if options & 4:
                    flags |= re.MULTILINE
            attributes = self.read_attributes()
            if sub_token in (TYPE_STRING, TYPE_REGEXP):
                encoding = self._get_encoding(attributes)
                try:
                    result = result.decode(encoding)
                except UnicodeDecodeError:
                    result = result.decode("unicode-escape")
            # string instance attributes are discarded
            if attributes and sub_token == TYPE_STRING:
                result = RubyString(result, attributes)
            if sub_token == TYPE_REGEXP:
                result = re.compile(str(result), flags)
            elif attributes:
                result.set_attributes(attributes)
        elif token == TYPE_STRING:
            size = self.read_long()
            result = self.fd.read(size)
        elif token == TYPE_SYMBOL:
            result = self.read_symreal()
        elif token == TYPE_FIXNUM:
            result = self.read_long()
        elif token == TYPE_ARRAY:
            num_elements = self.read_long()
            # noinspection PyUnusedLocal
            result = [self.read() for x in range(num_elements)]
        elif token == TYPE_HASH:
            num_elements = self.read_long()
            result = {}
            for x in range(num_elements):
                key = self.read()
                value = self.read()
                result[key] = value
            result = result
        elif token == TYPE_FLOAT:
            size = self.read_long()
            floatn = self.fd.read(size)
            floatn = floatn.split(b"\0")
            result = float(floatn[0].decode("utf-8"))
        elif token == TYPE_BIGNUM:
            sign = 1 if self.fd.read(1) == b"+" else -1
            num_elements = self.read_long()
            result = 0
            factor = 1
            for x in range(num_elements):
                result += self.read_short() * factor
                factor *= 2**16
            result *= sign
        elif token == TYPE_REGEXP:
            size = self.read_long()
            result = self.fd.read(size)
        elif token == TYPE_USRMARSHAL:
            class_symbol = self.read()
            if not isinstance(class_symbol, Symbol):
                raise ValueError("invalid class name: %r" % class_symbol)
            class_name = class_symbol.name
            attr_list = self.read()
            python_class = self.registry.get(class_name, UsrMarshal)
            if not issubclass(python_class, UsrMarshal):
                raise ValueError(
                    "invalid class mapping for %r: %r should be a subclass of %r."
                    % (class_name, python_class, UsrMarshal)
                )
            result = python_class(class_name)
            result.marshal_load(attr_list)
        elif token == TYPE_SYMLINK:
            result = self.read_symlink()
        elif token == TYPE_LINK:
            link_id = self.read_long()
            if object_index and link_id >= object_index:
                raise ValueError(
                    "invalid link destination: %d should be lower than %d."
                    % (link_id, object_index)
                )
            result = self.objects[link_id]
        elif token == TYPE_USERDEF:
            class_symbol = self.read()
            private_data = self.read(TYPE_STRING)
            if not isinstance(class_symbol, Symbol):
                raise ValueError("invalid class name: %r" % class_symbol)
            class_name = class_symbol.name
            python_class = self.registry.get(class_name, UserDef)
            if not issubclass(python_class, UserDef):
                raise ValueError(
                    "invalid class mapping for %r: %r should be a subclass of %r."
                    % (class_name, python_class, UserDef)
                )
            result = python_class(class_name)
            # noinspection PyProtectedMember
            result._load(private_data)
        elif token == TYPE_MODULE:
            data = self.read(TYPE_STRING)
            module_name = data.decode()
            result = Module(module_name, None)
        elif token == TYPE_OBJECT:
            class_symbol = self.read()
            assert isinstance(class_symbol, Symbol)
            class_name = class_symbol.name
            python_class = self.registry.get(class_name, RubyObject)
            if not issubclass(python_class, RubyObject):
                raise ValueError(
                    "invalid class mapping for %r: %r should be a subclass of %r."
                    % (class_name, python_class, RubyObject)
                )
            attributes = self.read_attributes()
            result = python_class(class_name, attributes)
        elif token == TYPE_EXTENDED:
            class_name = self.read(TYPE_STRING)
            result = Extended(class_name, None)
        elif token == TYPE_CLASS:
            data = self.read(TYPE_STRING)
            class_name = data.decode()
            if class_name in self.registry:
                result = self.registry[class_name]
            else:
                result = type(
                    class_name.rpartition(":")[2],
                    (RubyObject,),
                    {"ruby_class_name": class_name},
                )
        else:
            raise ValueError("token %s is not recognized" % token)
        if object_index is not None:
            self.objects[object_index - 1] = result
        return result

    @staticmethod
    def _get_encoding(attrs):
        encoding = "latin1"
        if attrs.get("E") is True:
            encoding = "utf-8"
        elif "encoding" in attrs:
            encoding = attrs["encoding"].decode()
        return encoding

    def read_attributes(self):
        attr_count = self.read_long()
        attrs = {}
        for x in range(attr_count):
            attr_name = self.read()
            attr_value = self.read()
            attrs[attr_name.name] = attr_value
        return attrs

    def read_short(self):
        return read_ushort(self.fd)

    def read_long(self):
        length = read_sbyte(self.fd)
        if length == 0:
            return 0
        if 5 < length < 128:
            return length - 5
        elif -129 < length < -5:
            return length + 5
        result = 0
        factor = 1
        for s in range(abs(length)):
            result += read_ubyte(self.fd) * factor
            factor *= 256
        if length < 0:
            result = result - factor
        return result

    def read_symbol(self):
        ivar = 0
        while True:
            token = self.fd.read(1)
            if token == TYPE_IVAR:
                ivar = 1
                continue
            elif token == TYPE_SYMBOL:
                return self.read_symreal()
            elif token == TYPE_SYMLINK:
                if ivar:
                    raise ValueError("dump format error (symlink with encoding)")
                return self.read_symlink()
            raise ValueError("error while reading symbol with token %r" % token)

    def read_symlink(self):
        symlink_id = self.read_long()
        return self.symbols[symlink_id]

    def read_symreal(self):
        size = self.read_long()
        result = self.fd.read(size)
        result = Symbol(result.decode("utf-8"))
        self.symbols.append(result)
        return result


def load(fd, registry=None):
    assert fd.read(1) == b"\x04"
    assert fd.read(1) == b"\x08"

    loader = Reader(fd, registry=registry)
    return loader.read()


def loads(byte_text, registry=None):
    return load(io.BytesIO(byte_text), registry=registry)
